"""
Artifact utils module.
"""
from __future__ import annotations

import typing
from hashlib import sha1
from mimetypes import guess_type
from pathlib import Path

if typing.TYPE_CHECKING:
    from digitalhub_core.entities.artifacts.entity import Artifact


def get_artifact_info(artifact: Artifact) -> dict:
    """
    Get the information of an artifact.

    Parameters
    ----------
    artifact : Artifact
        The artifact.

    Returns
    -------
    dict
        The information of the artifact.
    """
    return {
        "id": f"store://{artifact.project}/artifacts/{artifact.kind}/{artifact.name}:{artifact.id}",
        "key": artifact.name,
        "kind": artifact.kind,
    }


def calculate_blob_hash(data_path: str) -> str:
    """
    Calculate the hash of a file.

    Parameters
    ----------
    data_path : str
        Path to the file.

    Returns
    -------
    str
        The hash of the file.
    """
    with open(data_path, "rb") as f:
        data = f.read()
        return f"sha1_{sha1(data).hexdigest()}"


def get_file_size(data_path: str) -> int:
    """
    Get the size of a file.

    Parameters
    ----------
    data_path : str
        Path to the file.

    Returns
    -------
    int
        The size of the file.
    """
    return Path(data_path).stat().st_size


def get_file_mime_type(data_path: str) -> str:
    """
    Get the mime type of a file.

    Parameters
    ----------
    data_path : str
        Path to the file.

    Returns
    -------
    str
        The mime type of the file.
    """
    return guess_type(data_path)[0]


def get_file_extension(data_path: str) -> str:
    """
    Get the extension of a file.

    Parameters
    ----------
    data_path : str
        Path to the file.

    Returns
    -------
    str
        The extension of the file.
    """
    return Path(data_path).suffix[1:]



"""
Models for task specifications.
"""
from __future__ import annotations

from pydantic import BaseModel
from typing_extensions import Literal


class Volume(BaseModel):
    """
    Volume model.
    """

    volume_type: Literal["config_map", "secret", "persistent_volume_claim"]
    """Volume type."""

    name: str
    """Volume name."""

    mount_path: str
    """Volume mount path."""


class NodeSelector(BaseModel):
    """
    NodeSelector model.
    """

    key: str
    """Node selector key."""

    value: str
    """Node selector value."""


class Resource(BaseModel):
    """
    Resource model.
    """

    resource_type: Literal["cpu", "memory", "gpu"]
    """Resource kind (cpu, memory, gpu)."""

    requests: int
    """Resource requests."""

    limits: int
    """Resource limits."""


class Env(BaseModel):
    """
    Env variable model.
    """

    name: str
    """Env variable name."""

    value: str
    """Env variable value."""


class K8sResources(BaseModel):
    """
    K8sResources model.
    """

    node_selector: NodeSelector = {}
    """Node selector."""

    volumes: list[Volume] = []
    """List of volumes."""

    resources: Resource = {}
    """Resources restrictions."""

    env: list[Env] = []
    """Env variables."""



"""
Module for performing operations on tasks.
"""
from __future__ import annotations

import typing

from digitalhub_core.context.builder import check_context, get_context
from digitalhub_core.entities.tasks.entity import task_from_dict, task_from_parameters
from digitalhub_core.utils.api import api_base_delete, api_base_read, api_base_update
from digitalhub_core.utils.io_utils import read_yaml

if typing.TYPE_CHECKING:
    from digitalhub_core.entities.tasks.entity import Task


def create_task(**kwargs) -> Task:
    """
    Create a new object instance.

    Parameters
    ----------
    **kwargs
        Keyword arguments.

    Returns
    -------
    Task
       Object instance.
    """
    return task_from_parameters(**kwargs)


def create_task_from_dict(obj: dict) -> Task:
    """
    Create a new object instance.

    Parameters
    ----------
    obj : dict
        Object dictionary.

    Returns
    -------
    Task
       Object instance.
    """
    return task_from_dict(obj)


def new_task(
    project: str,
    kind: str,
    uuid: str | None = None,
    source: str | None = None,
    labels: list[str] | None = None,
    function: str | None = "",
    node_selector: dict | None = None,
    volumes: list[dict] | None = None,
    resources: dict | None = None,
    env: list[dict] | None = None,
    **kwargs,
) -> Task:
    """
    Create a new object instance.

    Parameters
    ----------
    project : str
        Name of the project.
    kind : str
        The type of the task.
    uuid : str
        UUID.
    source : str
        Remote git source for object.
    labels : list[str]
        List of labels.
    function : str
        The function string identifying the function.
    node_selector : dict
        The node selector of the task.
    volumes : list[dict]
        The volumes of the task.
    resources : dict
        Kubernetes resources for the task.
    env : list[dict]
        The env variables of the task.
    **kwargs
        Spec keyword arguments.

    Returns
    -------
    Task
       Object instance.
    """
    obj = create_task(
        project=project,
        kind=kind,
        uuid=uuid,
        source=source,
        labels=labels,
        function=function,
        node_selector=node_selector,
        volumes=volumes,
        resources=resources,
        env=env,
        **kwargs,
    )
    obj.save()
    return obj


def get_task(project: str, name: str) -> Task:
    """
    Get object from backend.

    Parameters
    ----------
    project : str
        Name of the project.
    name : str
        The name of the task.

    Returns
    -------
    Task
        Object instance.
    """
    api = api_base_read("tasks", name)
    obj = get_context(project).read_object(api)
    return create_task_from_dict(obj)


def import_task(file: str) -> Task:
    """
    Get object from file.

    Parameters
    ----------
    file : str
        Path to the file.

    Returns
    -------
    Task
        Object instance.
    """
    obj = read_yaml(file)
    check_context(obj.get("project"))
    return create_task_from_dict(obj)


def delete_task(project: str, name: str, cascade: bool = True) -> dict:
    """
    Delete task from the backend.

    Parameters
    ----------
    project : str
        Name of the project.
    name : str
        The name of the task.
    cascade : bool
        Whether to cascade delete.

    Returns
    -------
    dict
        Response from backend.
    """
    api = api_base_delete("tasks", name, cascade=cascade)
    return get_context(project).delete_object(api)


def update_task(task: Task) -> dict:
    """
    Update task.

    Parameters
    ----------
    task : Task
        The task object to update.

    Returns
    -------
    dict
        Response from backend.
    """
    api = api_base_update("tasks", task.id)
    return get_context(task.project).update_object(task.to_dict(), api)



"""
Common IO utils.
"""
from __future__ import annotations

from pathlib import Path

import yaml

####################
# Writers
####################


def write_yaml(filepath: str | Path, obj: dict) -> None:
    """
    Write a dict to a yaml file.

    Parameters
    ----------
    filepath : str | Path
        The yaml file path to write.
    obj : dict
        The dict to write.

    Returns
    -------
    None
    """
    with open(filepath, "w", encoding="utf-8") as out_file:
        yaml.dump(obj, out_file, sort_keys=False)


####################
# Readers
####################


def read_yaml(filepath: str | Path) -> dict:
    """
    Read a yaml file and return a dict.

    Parameters
    ----------
    filepath : str | Path
        The yaml file path to read.

    Returns
    -------
    dict
        The yaml file content.
    """
    with open(filepath, "r", encoding="utf-8") as in_file:
        data = yaml.load(in_file, Loader=yaml.SafeLoader)
    return data


def read_text(filepath: str | Path) -> str:
    """
    Read a file and return the text.

    Parameters
    ----------
    filepath : str | Path
        The file path to read.

    Returns
    -------
    str
        The file content.
    """
    return Path(filepath).read_text(encoding="utf-8")



"""
Entity metadata module.
"""
from __future__ import annotations

from digitalhub_core.entities._base.base import ModelObj


class Metadata(ModelObj):
    """
    A class representing the metadata of an entity.
    """

    def __init__(
        self,
        name: str | None = None,
        source: str | None = None,
        labels: list[str] | None = None,
        embedded: bool = True,
        created: str | None = str,
        updated: str | None = str,
    ) -> None:
        """
        Constructor.

        Parameters
        ----------
        name : str
            Name or UUID of the object.
        source : str
            Source of the entity.
        labels : list[str]
            A list of labels.
        embedded : bool
            Whether the entity is embedded.
        created : str
            Created date.
        updated : str
            Updated date.
        """
        self.name = name
        self.source = source
        self.labels = labels
        self.embedded = embedded
        self.created = created
        self.updated = updated

    @classmethod
    def from_dict(cls, obj: dict) -> "Metadata":
        """
        Return entity metadata object from dictionary.

        Parameters
        ----------
        obj : dict
            A dictionary containing the attributes of the entity metadata.

        Returns
        -------
        Metadata
            An entity metadata object.
        """
        return cls(**obj)



"""
Task specification module.
"""
from __future__ import annotations

from digitalhub_core.entities._base.spec import Spec, SpecParams
from digitalhub_core.entities.tasks.models import Env, NodeSelector, Resource, Volume


class TaskSpec(Spec):
    """Task specification."""

    def __init__(
        self,
        function: str,
        node_selector: NodeSelector | None = None,
        volumes: list[Volume] | None = None,
        resources: Resource | None = None,
        env: list[Env] | None = None,
        **kwargs,
    ) -> None:
        """
        Constructor.

        Parameters
        ----------
        function : str
            The function string of the task.
        node_selector : NodeSelector
            The node selector of the task.
        volumes : list[Volume]
            The volumes of the task.
        resources : Resource
            Kubernetes resources for the task.
        env : list[Env]
            The env variables of the task.
        """
        self.function = function
        self.node_selector = node_selector
        self.volumes = volumes
        self.resources = resources
        self.env = env

        self._any_setter(**kwargs)


class TaskParams(SpecParams):
    """
    Base task model.
    """

    function: str
    """Function string."""

    node_selector: NodeSelector = None
    """Node selector."""

    volumes: list[Volume] = None
    """List of volumes."""

    resources: Resource = None
    """Resources restrictions."""

    env: list[Env] = None
    """Env variables."""



"""
Metadata factory module.
"""
from __future__ import annotations

import typing

from digitalhub_core.utils.generic_utils import get_timestamp

if typing.TYPE_CHECKING:
    from digitalhub_core.entities._base.metadata import Metadata


def build_metadata(metadata_cls: Metadata, **kwargs) -> Metadata:
    """
    Build entity metadata object.

    Parameters
    ----------
    metadata_cls: Metadata
        Metadata object.
    **kwargs
        Keyword arguments.

    Returns
    -------
    Metadata
        An entity metadata object.
    """
    kwargs = parse_arguments(**kwargs)
    return metadata_cls(**kwargs)


def parse_arguments(**kwargs) -> dict:
    """
    Parse keyword arguments and add default values.

    Parameters
    ----------
    **kwargs
        Keyword arguments.

    Returns
    -------
    dict
        A dictionary containing the entity metadata attributes.
    """
    if "created" not in kwargs or kwargs["created"] is None:
        kwargs["created"] = get_timestamp()
    if "updated" not in kwargs or kwargs["updated"] is None:
        kwargs["updated"] = kwargs["created"]
    return kwargs



"""
Run module.
"""
from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    from digitalhub_core.entities.artifacts.entity import Artifact


class RunResults:
    """
    A class representing a run results.
    """

    def __init__(
        self,
        artifacts: list[Artifact] | None = None,
    ) -> None:
        """
        Constructor.

        Parameters
        ----------
        artifacts : list[Artifact]
            The artifacts.
        """
        self.artifacts = artifacts



"""
FunctionStatus class module.
"""
from __future__ import annotations

from digitalhub_core.entities._base.status import Status


class FunctionStatus(Status):
    """
    Status class for function entities.
    """

    def __init__(self, state: str, message: str | None = None, tasks: list[dict] | None = None) -> None:
        """
        Constructor.

        Parameters
        ----------
        tasks : list[dict]
            List of tasks associated to function.
        """
        super().__init__(state, message)
        self.tasks = {}
        self._set_tasks(tasks)

    def _set_tasks(self, tasks: list[dict] | None) -> None:
        """
        Set tasks.

        Parameters
        ----------
        tasks : list[dict]
            List of tasks associated to function.

        Returns
        -------
        None
        """
        if tasks is None:
            return
        try:
            self.tasks = {t["kind"]: t for t in tasks}
        except KeyError:
            raise ValueError("Malformed tasks.")



"""
Common URI utils.
"""
from __future__ import annotations

from urllib.parse import urlparse


def map_uri_scheme(uri: str) -> str:
    """
    Map an URI scheme to a common scheme.

    Parameters
    ----------
    uri : str
        URI.

    Returns
    -------
    str
        Mapped scheme type.

    Raises
    ------
    ValueError
        If the scheme is unknown.
    """
    scheme = urlparse(uri).scheme
    if scheme in ["", "file", "local"]:
        return "local"
    if scheme in ["http", "https", "remote"]:
        return "remote"
    if scheme in ["s3", "s3a", "s3n"]:
        return "s3"
    if scheme in ["sql", "postgresql"]:
        return "sql"
    raise ValueError(f"Unknown scheme '{scheme}'!")



"""
Spec factory entity.
"""
from __future__ import annotations

import importlib
import importlib.metadata
import typing

if typing.TYPE_CHECKING:
    from digitalhub_core.entities._base.spec import Spec


def build_spec(
    entity: str,
    kind: str,
    layer_digitalhub: str | None = None,
    framework_runtime: str | None = None,
    validate: bool = True,
    **kwargs,
) -> Spec:
    """
    Build runtimes.

    Parameters
    ----------
    entity : str
        Type of entity.
    kind : str
        The type of Spec to build.
    layer_digitalhub : str
        Layer name.
    framework_runtime : str
        Framework name.
    validate : bool
        Validate arguments.

    Returns
    -------
    Spec
        Spec object.
    """
    if layer_digitalhub is None and framework_runtime is None:
        raise ValueError("Either layer or framework must be provided.")

    if framework_runtime is not None:
        # This could override layer
        layer_digitalhub = check_module_existence_by_framework(framework_runtime)
    else:
        check_layer_existence(layer_digitalhub)

    module_to_import = f"{layer_digitalhub}.entities.{entity}.spec"
    registry = import_registry(module_to_import)

    try:
        class_spec, class_params = registry[kind]
    except KeyError:
        raise ValueError(f"Spec {kind} not found in registry.")
    except ValueError as err:
        raise ValueError(f"Spec registry of module {module_to_import} is not valid: {err}.")

    # Validate arguments
    if validate:
        kwargs = class_params(**kwargs).dict()

    return class_spec(**kwargs)


def check_module_existence_by_framework(framework: str) -> str:
    """
    Try to import registry from implemented module.

    Parameters
    ----------
    framework : str
        Framework name.

    Returns
    -------
    str
        Valid module name.
    """
    # Cycle over digitalhub layers modules (data, ml, ai).
    for layer in ["data", "ml", "ai"]:
        # Try to import module
        module_name = f"digitalhub_{layer}_{framework}"
        try:
            importlib.metadata.distribution(module_name)
            return module_name
        except importlib.metadata.PackageNotFoundError:
            continue
    # If no module is found raise error
    else:
        raise ModuleNotFoundError(f"Module not found in digitalhub layers data, ml, ai for framework {framework}")


def check_layer_existence(layer: str) -> None:
    """
    Check if layer exists.

    Parameters
    ----------
    layer : str
        Layer name.

    Returns
    -------
    None
    """
    try:
        importlib.metadata.distribution(layer)
    except importlib.metadata.PackageNotFoundError:
        raise ModuleNotFoundError(f"Layer {layer} not found.")


def import_registry(module_to_import: str) -> dict[str, list[Spec, Spec]]:
    """
    Import registry from implemented module.

    Parameters
    ----------
    module_to_import : str
        Module name.

    Returns
    -------
    typing.Any
        Registry.
    """
    module = importlib.import_module(module_to_import)
    return getattr(module, "SPEC_REGISTRY")



"""
Abstract entity module.
"""
from __future__ import annotations

from abc import ABCMeta, abstractmethod

from digitalhub_core.entities._base.base import ModelObj


class Entity(ModelObj, metaclass=ABCMeta):
    """
    Abstract class for entities.
    """

    # Attributes to render as dict. Need to be expanded in subclasses.
    _obj_attr = ["kind", "metadata", "spec", "status"]

    @abstractmethod
    def save(self, update: bool = False) -> dict:
        """
        Abstract save method.
        """

    @abstractmethod
    def export(self, filename: str | None = None) -> None:
        """
        Abstract save method.
        """

    def to_dict(
        self,
        include_all_non_private: bool = False,
    ) -> dict:
        """
        Return object as dict with all keys.

        Parameters
        ----------
        include_all_non_private : bool
            Whether to include all non-private attributes. If False, only
            attributes in the _obj_attr list will be included.

        Returns
        -------
        dict
            A dictionary containing the attributes of the entity instance.
        """
        dict_ = super().to_dict()
        if include_all_non_private:
            return dict_
        return {k: v for k, v in dict_.items() if k in self._obj_attr}

    @classmethod
    def from_dict(
        cls,
        obj: dict,
        validate: bool = True,
    ) -> "Entity":
        """
        Create object instance from a dictionary.

        Parameters
        ----------
        obj : dict
            Dictionary to create object from.
        validate : bool
            Flag to indicate if arguments validation must be ignored.

        Returns
        -------
        Self
            Self instance.
        """
        parsed_dict = cls._parse_dict(
            obj,
            validate=validate,
        )
        return cls(**parsed_dict)

    @staticmethod
    @abstractmethod
    def _parse_dict(
        obj: dict,
        validate: bool = True,
    ) -> dict:
        """
        Get dictionary and parse it to a valid entity dictionary.
        """

    def __repr__(self) -> str:
        """
        Return string representation of the entity object.

        Returns
        -------
        str
            A string representing the entity instance.
        """
        return str(self.to_dict(include_all_non_private=True))



"""
Base Runtime module.
"""
from __future__ import annotations

import typing
from abc import abstractmethod
from typing import Any, Callable

from digitalhub_core.utils.exceptions import EntityError
from digitalhub_core.utils.logger import LOGGER

if typing.TYPE_CHECKING:
    from digitalhub_core.runtimes.results import RunResults


class Runtime:
    """
    Base Runtime class.
    """

    ##################################
    # Abstract methods
    ##################################

    allowed_actions = []

    @abstractmethod
    def build(self, function: dict, task: dict, run: dict) -> dict:
        """
        Build run spec.
        """

    @abstractmethod
    def run(self, run: dict) -> dict:
        """
        Execute run task.
        """

    @staticmethod
    @abstractmethod
    def _get_function(action: str) -> Callable:
        """
        Get function from action.
        """

    @staticmethod
    @abstractmethod
    def results(run_status: dict) -> RunResults:
        """
        Get run results.
        """

    ##################################
    # Private methods
    ##################################

    def _validate_task(self, run: dict) -> str:
        """
        Validate task.

        Parameters
        ----------
        run : dict
            Run object dictionary.

        Returns
        -------
        str
            Action to execute.
        """
        action = self._get_action(run)
        if action not in self.allowed_actions:
            msg = f"Task {action} not allowed for {self.__class__.__name__} runtime."
            LOGGER.error(msg)
            raise EntityError(msg)
        return action

    @staticmethod
    def _get_action(run: dict) -> str:
        """
        Get action to execute from run.

        The run object stores in its spec the identifier of the task
        it is associated with. The task string is derived from the
        function string, and has the following format:

        <function-kind>+<task-action>://<function-name>:<function-id>

        Parameters
        ----------
        run : dict
            Run object dictionary.

        Returns
        -------
        str
            Action to execute.

        Raises
        ------
        RuntimeError
            If malformed run spec.

        Examples
        --------
        >>> run = {"spec": {"task": dbt+transform://dbt-example-function:some-uuid4}}
        >>> Runtime._get_action(run)
        'transform'
        """
        try:
            return run["spec"]["task"].split(":")[0].split("+")[1]
        except (KeyError, IndexError):
            msg = "Malformed run spec."
            LOGGER.exception(msg)
            raise RuntimeError(msg)

    @staticmethod
    def _execute(func: Callable, *args, **kwargs) -> Any:
        """
        Execute function.

        Parameters
        ----------
        func : Callable
            Function to execute.
        *args
            Function arguments.
        **kwargs
            Function keyword arguments.

        Returns
        -------
        Any
            Function result.
        """
        try:
            return func(*args, **kwargs)
        except Exception:
            msg = "Something got wrong during function execution."
            LOGGER.exception(msg)
            raise RuntimeError(msg)



"""
Entity specification module.
"""
from __future__ import annotations

from digitalhub_core.entities._base.base import ModelObj
from pydantic import BaseModel


class Spec(ModelObj):
    """
    A class representing the specification of an entity.
    """

    @classmethod
    def from_dict(cls, obj: dict) -> "Spec":
        """
        Return entity specification object from dictionary.

        Parameters
        ----------
        obj : dict
            A dictionary containing the attributes of the entity specification.

        Returns
        -------
        EntitySpec
            An entity specification object.
        """
        return cls(**obj)


class SpecParams(BaseModel):
    """
    A class representing the parameters of an entity.
    """



"""
Run module.
"""
from __future__ import annotations

import typing
from collections import namedtuple
from pathlib import Path

from digitalhub_core.context.builder import get_context
from digitalhub_core.entities._base.entity import Entity
from digitalhub_core.entities._base.status import State
from digitalhub_core.entities._builders.metadata import build_metadata
from digitalhub_core.entities._builders.spec import build_spec
from digitalhub_core.entities._builders.status import build_status
from digitalhub_core.entities.runs.metadata import RunMetadata
from digitalhub_core.entities.runs.status import RunStatus
from digitalhub_core.runtimes.builder import build_runtime
from digitalhub_core.utils.api import api_base_create, api_base_read, api_base_update, api_ctx_read
from digitalhub_core.utils.exceptions import EntityError
from digitalhub_core.utils.generic_utils import build_uuid, get_timestamp
from digitalhub_core.utils.io_utils import write_yaml

if typing.TYPE_CHECKING:
    from digitalhub_core.context.context import Context
    from digitalhub_core.entities.runs.spec import RunSpec
    from digitalhub_core.runtimes.base import Runtime


TaskString = namedtuple(
    "TaskString",
    ["function_kind", "task_kind", "function_name", "function_id", "task_id"],
)


class Run(Entity):
    """
    A class representing a run.
    """

    def __init__(
        self,
        project: str,
        uuid: str,
        kind: str,
        metadata: RunMetadata,
        spec: RunSpec,
        status: RunStatus,
    ) -> None:
        """
        Constructor.

        Parameters
        ----------
        project : str
            Name of the project.
        uuid : str
            UUID.
        kind : str
            The type of the run.
        metadata : RunMetadata
            Metadata of the object.
        spec : RunSpec
            Specification of the object.
        status : RunStatus
            Status of the object.
        """
        super().__init__()
        self.project = project
        self.id = uuid
        self.kind = kind
        self.metadata = metadata
        self.spec = spec
        self.status = status

        # Add attributes to be used in the to_dict method
        self._obj_attr.extend(["project", "id"])

    #############################
    #  Save / Export
    #############################

    def save(self, update: bool = False) -> dict:
        """
        Save run into backend.

        Parameters
        ----------
        uuid : str
            Ignore this parameter.

        Returns
        -------
        dict
            Mapping representation of Run from backend.
        """
        obj = self.to_dict(include_all_non_private=True)

        if not update:
            api = api_base_create("runs")
            return self._context().create_object(obj, api)

        self.metadata.updated = obj["metadata"]["updated"] = get_timestamp()
        api = api_base_update("runs", self.id)
        return self._context().update_object(obj, api)

    def export(self, filename: str | None = None) -> None:
        """
        Export object as a YAML file.

        Parameters
        ----------
        filename : str
            Name of the export YAML file. If not specified, the default value is used.

        Returns
        -------
        None
        """
        obj = self.to_dict()
        if filename is None:
            filename = f"{self.kind}_{self.name}_{self.id}.yml"
        pth = Path(self.project) / filename
        pth.parent.mkdir(parents=True, exist_ok=True)
        write_yaml(pth, obj)

    #############################
    #  Context
    #############################

    def _context(self) -> Context:
        """
        Get context.

        Returns
        -------
        Context
            Context.
        """
        return get_context(self.project)

    #############################
    #  Run Methods
    #############################

    def build(self) -> None:
        """
        Build run.

        Returns
        -------
        None
        """
        function = self._get_function()
        task = self._get_task()
        runtime = self._get_runtime()
        new_spec = runtime.build(function, task, self.to_dict())
        self.spec = build_spec("runs", self.kind, layer_digitalhub="digitalhub_core", validate=False, **new_spec)
        self._set_status({"state": State.BUILT.value})
        self.save()

    def run(self) -> Run:
        """
        Run run.

        Returns
        -------
        Run
            Run object.
        """
        if self.spec.local_execution:
            if not self.status.state == State.BUILT.value:
                raise EntityError("Run is not in built state. Build it again.")
            self._set_status({"state": State.RUNNING.value})
            self.save(update=True)

        runtime = self._get_runtime()
        try:
            status = runtime.run(self.to_dict(include_all_non_private=True))
        except Exception as err:
            status = {"state": State.ERROR.value, "message": str(err)}
        self._set_status(status)
        self.save(update=True)
        return self

    def results(self) -> list[object]:
        """
        Get run objects results.

        Returns
        -------
        dict
            Results from backend.
        """
        runtime = self._get_runtime()
        return runtime.results(self.status.to_dict())

    def refresh(self) -> Run:
        """
        Get run from backend.

        Returns
        -------
        Run
            Run object.
        """
        api = api_base_read("runs", self.id)
        obj = self._context().read_object(api)
        refreshed_run = self.from_dict(obj)
        self.kind = refreshed_run.kind
        self.metadata = refreshed_run.metadata
        self.spec = refreshed_run.spec
        self.status = refreshed_run.status
        return self

    def logs(self) -> dict:
        """
        Get run's logs from backend.
        Returns empty dictionary if context is local.

        Returns
        -------
        dict
            Logs from backend.
        """
        if self._context().local:
            return {}
        api = api_base_read("runs", self.id) + "/log"
        return self._context().read_object(api)

    def _set_status(self, status: dict) -> None:
        """
        Set run status.

        Parameters
        ----------
        status : dict
            Status to set.

        Returns
        -------
        None

        Raises
        ------
        EntityError
            If status is not a dictionary or a Status object.
        """
        if not isinstance(status, dict):
            raise EntityError("Status must be a dictionary.")
        self.status: RunStatus = build_status(RunStatus, **status)

    #############################
    #  Functions and Tasks
    #############################

    def _parse_task_string(self) -> TaskString:
        """
        Parse task string.

        Returns
        -------
        None
        """
        kinds, func = self.spec.task.split("://")
        fnc_kind, tsk_kind = kinds.split("+")
        fnc_name, fnc_id = func.split("/")[1].split(":")
        return TaskString(fnc_kind, tsk_kind, fnc_name, fnc_id, self.spec.task_id)

    def _get_function(self) -> dict:
        """
        Get function from backend. Reimplemented to avoid circular imports.

        Returns
        -------
        dict
            Function from backend.
        """
        parsed = self._parse_task_string()
        api = api_ctx_read(self.project, "functions", parsed.function_name, parsed.function_id)
        return self._context().read_object(api)

    def _get_task(self) -> dict:
        """
        Get task from backend. Reimplemented to avoid circular imports.

        Returns
        -------
        dict
            Task from backend.
        """
        parsed = self._parse_task_string()
        api = api_base_read("tasks", parsed.task_id)
        return self._context().read_object(api)

    #############################
    # Runtimes
    #############################

    def _get_runtime(self) -> Runtime:
        """
        Build runtime to build run or execute it.

        Returns
        -------
        Runtime
            Runtime object.
        """
        fnc_kind = self._parse_task_string().function_kind
        return build_runtime(fnc_kind)

    #############################
    #  Static interface methods
    #############################

    @staticmethod
    def _parse_dict(
        obj: dict,
        validate: bool = True,
    ) -> dict:
        """
        Get dictionary and parse it to a valid entity dictionary.

        Parameters
        ----------
        entity : str
            Entity type.
        obj : dict
            Dictionary to parse.

        Returns
        -------
        dict
            A dictionary containing the attributes of the entity instance.
        """
        project = obj.get("project")
        kind = obj.get("kind")
        uuid = build_uuid(obj.get("id"))
        metadata = build_metadata(RunMetadata, **obj.get("metadata", {}))
        spec = build_spec(
            "runs",
            kind,
            layer_digitalhub="digitalhub_core",
            validate=validate,
            **obj.get("spec", {}),
        )
        status = build_status(RunStatus, **obj.get("status", {}))
        return {
            "project": project,
            "uuid": uuid,
            "kind": kind,
            "metadata": metadata,
            "spec": spec,
            "status": status,
        }


def run_from_parameters(
    project: str,
    task: str,
    task_id: str,
    kind: str,
    uuid: str | None = None,
    source: str | None = None,
    labels: list[str] | None = None,
    inputs: dict | None = None,
    outputs: list | None = None,
    parameters: dict | None = None,
    local_execution: bool = False,
    **kwargs,
) -> Run:
    """
    Create run.

    Parameters
    ----------
    project : str
        Name of the project.
    task_id : str
        Identifier of the task associated with the run.
    task : str
        Name of the task associated with the run.
    kind : str
        The type of the run.
    uuid : str
        UUID.
    source : str
        Remote git source for object.
    labels : list[str]
        List of labels.
    inputs : dict
        Inputs of the run.
    outputs : list
        Outputs of the run.
    parameters : dict
        Parameters of the run.
    local_execution : bool
        Flag to determine if object has local execution.
    embedded : bool
        Flag to determine if object must be embedded in project.
    **kwargs
        Spec keyword arguments.

    Returns
    -------
    Run
        Run object.
    """
    uuid = build_uuid(uuid)
    metadata = build_metadata(
        RunMetadata,
        project=project,
        name=uuid,
        source=source,
        labels=labels,
    )
    spec = build_spec(
        "runs",
        kind,
        layer_digitalhub="digitalhub_core",
        task=task,
        task_id=task_id,
        inputs=inputs,
        outputs=outputs,
        parameters=parameters,
        local_execution=local_execution,
        **kwargs,
    )
    status = build_status(RunStatus)
    return Run(
        project=project,
        uuid=uuid,
        kind=kind,
        metadata=metadata,
        spec=spec,
        status=status,
    )


def run_from_dict(obj: dict) -> Run:
    """
    Create run from dictionary.

    Parameters
    ----------
    obj : dict
        Dictionary to create run from.

    Returns
    -------
    Run
        Run object.
    """
    return Run.from_dict(obj)



"""
DHCore Client module.
"""
from __future__ import annotations

import os

import requests
from digitalhub_core.client.objects.base import Client
from digitalhub_core.utils.exceptions import BackendError


class ClientDHCore(Client):
    """
    DHCore client.

    The DHCore client is used to communicate with the Digitalhub Core backendAPI via REST.
    At creation, the client trys to get the endpoint and authentication parameters
    from the environment variables. In case the endpoint is not set, it raises an exception.
    """

    def __init__(self) -> None:
        """
        Constructor.
        """
        super().__init__()
        self._endpoint = self._get_endpoint()
        self._auth = self._get_auth()

    def create_object(self, obj: dict, api: str) -> dict:
        """
        Create an object.

        Parameters
        ----------
        obj : dict
            The object to create.
        api : str
            The api to create the object with.

        Returns
        -------
        dict
            The created object.
        """
        return self._call("POST", api, json=obj)

    def read_object(self, api: str) -> dict:
        """
        Get an object.

        Parameters
        ----------
        api : str
            The api to get the object with.

        Returns
        -------
        dict
            The object.
        """
        return self._call("GET", api)

    def update_object(self, obj: dict, api: str) -> dict:
        """
        Update an object.

        Parameters
        ----------
        obj : dict
            The object to update.
        api : str
            The api to update the object with.

        Returns
        -------
        dict
            The updated object.
        """
        return self._call("PUT", api, json=obj)

    def delete_object(self, api: str) -> dict:
        """
        Delete an object.

        Parameters
        ----------
        api : str
            The api to delete the object with.

        Returns
        -------
        dict
            A generic dictionary.
        """
        resp = self._call("DELETE", api)
        if isinstance(resp, bool):
            resp = {"deleted": resp}
        return resp

    def _call(self, call_type: str, api: str, **kwargs) -> dict:
        """
        Make a call to the DHCore API.
        Keyword arguments are passed to the session.request function.

        Parameters
        ----------
        call_type : str
            The type of call to make.
        api : str
            The api to call.
        **kwargs
            Keyword arguments.

        Returns
        -------
        dict
            The response object.
        """
        url = self._endpoint + api
        kwargs["auth"] = self._auth
        response = None
        try:
            response = requests.request(call_type, url, timeout=60, **kwargs)
            response.raise_for_status()
            return response.json()
        except requests.exceptions.RequestException as e:
            if isinstance(e, requests.exceptions.Timeout):
                msg = "Request to DHCore backend timed out."
            elif isinstance(e, requests.exceptions.ConnectionError):
                msg = "Unable to connect to DHCore backend."
            else:
                msg = f"Backend error: {e}"
            raise BackendError(msg) from e

    @staticmethod
    def _get_endpoint() -> str:
        """
        Get DHub Core endpoint environment variables.

        Returns
        -------
        str
            DHub Core endpoint environment variables.

        Raises
        ------
        Exception
            If the endpoint of DHCore is not set in the env variables.
        """
        endpoint = os.getenv("DIGITALHUB_CORE_ENDPOINT")
        if endpoint is None:
            raise BackendError("Endpoint not set as environment variables.")
        if endpoint.endswith("/"):
            endpoint = endpoint[:-1]
        return endpoint

    @staticmethod
    def _get_auth() -> tuple[str, str] | None:
        """
        Get authentication parameters from the config.

        Returns
        -------
        tuple[str, str]
            The authentication parameters.
        """
        user = os.getenv("DIGITALHUB_CORE_USER")
        password = os.getenv("DIGITALHUB_CORE_PASSWORD")
        if user is None or password is None:
            return None
        return user, password

    @staticmethod
    def is_local() -> bool:
        """
        Declare if Client is local.

        Returns
        -------
        bool
            False
        """
        return False



"""
Local Client module.
"""
from copy import deepcopy

from digitalhub_core.client.objects.base import Client
from digitalhub_core.utils.exceptions import BackendError


class ClientLocal(Client):
    """
    Local client.

    The Local client can be used when a remote Digitalhub backend is not available.
    It handles the creation, reading, updating and deleting of objects in memory,
    storing them in a local dictionary.
    The functionality of the Local client is almost the same as the DHCore client.
    Main differences are:
        - Local client does delete objects on cascade.
        - The run execution are forced to be local.
    """

    def __init__(self) -> None:
        """
        Constructor.
        """
        super().__init__()
        self._db: dict[str, dict[str, dict]] = {}

    ########################
    # CRUD
    ########################

    def create_object(self, obj: dict, api: str) -> dict:
        """
        Create an object.

        Parameters
        ----------
        obj : dict
            The object to create.
        api : str
            The api to create the object with.

        Returns
        -------
        dict
            The created object.
        """
        project, dto, name, uuid, code = self._parse_api(api)
        try:
            # Check if dto is valid
            if dto is None:
                raise TypeError

            # Check if dto exists, if not create a mapping
            self._db.setdefault(dto, {})

            # Unversioned objects uses "base api". For example:
            #
            # POST /api/v1/projects
            # POST /api/v1/tasks
            # POST /api/v1/runs
            #
            # We do not have "name" attribute for tasks and runs
            # so we use the id to identify them. Projects has
            # the name attribute, not the id and is also not versioned,
            # so we use the name as storage key.
            if project is None:
                name = obj["name"] if dto == "projects" else obj["id"]
                if name in self._db[dto]:
                    code = 5
                    raise ValueError
                self._db[dto][name] = obj

            # Versioned objects uses "context api". For example:
            #
            # POST /api/v1/-/<project-name>/artifacts
            # POST /api/v1/-/<project-name>/functions
            # POST /api/v1/-/<project-name>/workflows
            #
            # We have bith "name" and "id" attributes for versioned objects
            # so we use them as storage keys. The "latest" key is used
            # to store the latest version of the object.
            else:
                name = obj["name"]
                uuid = obj["id"]
                self._db[dto].setdefault(name, {})
                self._db[dto][name][uuid] = obj
                self._db[dto][name]["latest"] = obj

            # Return the created object
            return obj

        # Key error are possibly raised by accessing invalid objects
        except (KeyError, TypeError):
            msg = self._format_msg(code, dto=dto)
            raise BackendError(msg)

        # If try to create already existing object
        except ValueError:
            msg = self._format_msg(code, dto=dto, name=name)
            raise BackendError(msg)

    def read_object(self, api: str) -> dict:
        """
        Get an object.

        Parameters
        ----------
        api : str
            The api to get the object with.

        Returns
        -------
        dict or None
            The object, or None if it doesn't exist.
        """
        project, dto, name, uuid, code = self._parse_api(api)
        try:
            # Unversioned objects
            # API examples
            #
            # GET /api/v1/projects/<name>
            #
            # self._parse_api() should return only dto

            if project is None:
                obj = self._db[dto][name]

                # If the object is a project, we need to add the project spec,
                # for example artifacts, functions, workflows, etc.
                if dto == "projects":
                    obj = self._get_project_spec(obj, name)

            # Versioned objects
            # API example
            #
            # GET /api/v1/-/<project-name>/artifacts/<uuid>
            #
            # self._parse_api() should return dto, name and uuid/version

            else:
                obj = self._db[dto][name][uuid]

            return obj

        except KeyError:
            msg = self._format_msg(code, project, dto, name, uuid)
            raise BackendError(msg)

    def update_object(self, obj: dict, api: str) -> dict:
        """
        Update an object.

        Parameters
        ----------
        obj : dict
            The object to update.
        api : str
            The api to update the object with.

        Returns
        -------
        dict
            The updated object.
        """
        project, dto, name, uuid, code = self._parse_api(api)
        try:
            # Unversioned objects
            # API example
            #
            # PUT /api/v1/projects/<name>

            if project is None:
                self._db[dto][name] = obj

            # Versioned objects
            # API example
            #
            # PUT /api/v1/-/<project-name>/artifacts/<uuid>

            else:
                self._db[dto][name][uuid] = obj

        except KeyError:
            msg = self._format_msg(code, project, dto, name, uuid)
            raise BackendError(msg)

        return obj

    def delete_object(self, api: str) -> dict:
        """
        Delete an object.

        Parameters
        ----------
        api : str
            The api to delete the object with.

        Returns
        -------
        dict
            A generic dictionary.
        """
        project, dto, name, uuid, code = self._parse_api(api)
        try:
            # Unversioned objects
            # API example
            #
            # DELETE /api/v1/projects/<name>

            if uuid is None:
                obj = self._db[dto].pop(name)

            # Versioned objects
            # API example
            #
            # DELETE /api/v1/-/<project-name>/artifacts/<uuid>
            #
            # We do not handle cascade in local client and
            # in the sdk we selectively delete objects by id,
            # not by name nor dto.

            else:
                obj = self._db[dto][name].pop(uuid)

        except KeyError:
            msg = self._format_msg(code, project, dto, name, uuid)
            raise BackendError(msg)
        return {"deleted": obj}

    ########################
    # Logic for CRUD
    ########################

    def _parse_api(self, api: str) -> list[str]:
        """
        Parse the given API.

        Parameters
        ----------
        api : str
            The API to parse.

        Returns
        -------
        list[str]
            The parsed API.
        """
        # Remove prefix from API
        api = api.removeprefix("/api/v1/")

        # Set context flag by default to False
        ctx = False

        # Remove context prefix from API and set context flag to True
        if api.startswith("-/"):
            ctx = True
            api = api[2:]

        # Remove delete flag from API. Local client does not handle cascade
        if api.endswith("?cascade=true") or api.endswith("?cascade=false"):
            api = api.removesuffix("?cascade=true").removesuffix("?cascade=false")

        # Return parsed elements
        return self._parse_api_elements(api, ctx)

    @staticmethod
    def _parse_api_elements(api: str, ctx: bool) -> tuple:
        """
        Parse the elements from the given API.
        Elements returned are: project-name, dto, name, uuid, error_code.

        Parameters
        ----------
        api : str
            The parsed API.
        ctx : bool
            True if the API is a context API.

        Returns
        -------
        tuple
            The parsed elements from the API.
        """
        # Split API path
        parsed = api.split("/")

        # Base API for versioned objects

        # POST /api/v1/<dto>
        # Returns None, dto, None, None
        if len(parsed) == 1:
            return None, parsed[0], None, None, 1

        # GET/DELETE/UPDATE /api/v1/<dto>/<name>
        # Return None, dto, name, None
        if len(parsed) == 2 and not ctx:
            return None, parsed[0], parsed[1], None, 2

        # Context API for versioned objects

        # POST /api/v1/-/<project>/<dto>
        # Returns project-name, dto, None, None
        if len(parsed) == 2 and ctx:
            return parsed[0], parsed[1], None, None, 1

        # GET/DELETE/UPDATE /api/v1/-/<project>/<dto>/<name>
        # Return project-name, dto, name, None
        if len(parsed) == 3:
            return parsed[0], parsed[1], parsed[2], None, 3

        # GET/DELETE/UPDATE /api/v1/-/<project>/<dto>/<name>/<uuid>
        # Return project-name, dto, name, uuid
        if len(parsed) == 4:
            return parsed[0], parsed[1], parsed[2], parsed[3], 4

    def _get_project_spec(self, obj: dict, name: str) -> dict:
        """
        Read the project spec.

        Parameters
        ----------
        obj : dict
            The project object.
        name : str
            The project name.

        Returns
        -------
        dict
            The project object with the spec.
        """
        # Deepcopy to avoid modifying the original object
        project = deepcopy(obj)
        spec = project.get("spec", {})

        # Get all entities associated with the project specs
        projects_entities = [k for k, _ in self._db.items() if k not in ["projects", "runs", "tasks"]]

        for entity_type in projects_entities:
            # Get all objects of the entity type for the project
            objs = self._db[entity_type]

            # Set empty list
            spec[entity_type] = []

            # Cycle through named objects
            for _, named_entities in objs.items():
                # Get latest version
                for version, entity in named_entities.items():
                    if version != "latest":
                        continue

                    # Deepcopy to avoid modifying the original object
                    copied = deepcopy(entity)

                    # Remove spec if not embedded
                    if not copied.get("metadata", {}).get("embedded", True):
                        copied.pop("spec", None)

                    # Add to project spec
                    if copied["project"] == name:
                        spec[entity_type].append(copied)

        return project

    ########################
    # Utils
    ########################

    @staticmethod
    def _format_msg(
        code: int,
        project: str = None,
        dto: str = None,
        name: str = None,
        uuid: str = None,
    ) -> str:
        """
        Format a message.

        Parameters
        ----------
        code : int
            The code to format the message with.
        project : str
            The project name.
        dto : str
            The DTO name.
        name : str
            The object name.
        uuid : str
            The object uuid.

        Returns
        -------
        str
            The formatted message.
        """
        msg = {
            1: f"Object '{dto}' to create is not valid",
            2: f"Object '{dto}' named '{name}' not found",
            3: f"Object '{dto}' named '{name}' for project '{project}' not found",
            4: f"Object '{dto}' named '{name}:{uuid}' for project '{project}' not found",
            5: f"Object '{dto}' named '{name}' already exists",
        }
        return msg[code]

    ########################
    # Interface methods
    ########################

    @staticmethod
    def is_local() -> bool:
        """
        Declare if Client is local.

        Returns
        -------
        bool
            True
        """
        return True
